----------------------------------------------------------------------------------
-- stripped down version in an attempt to get something to meet the timing constraint
--

-- Engineer:  			Naomi Jarvis, Carnegie Mellon University
-- Create Date:   	09:27:11 11/04/2013 
-- Module Name:   	findtime1 - Behavioral 
-- Project Name:  	CDC hit time finder for GlueX
-- Target Devices: 	FA125 spartan 6
-- Tool versions: 
-- Description: 
--
--   	external module searches adc data for a threshold crossing, then passes in NSAMPLES adc samples to this module
--	  	        the adc data passed in exceed threshold crossing at sample number XTHR_SAMP (first sample is numbered 0)
--              
--	this module     upsamples the adc data and then finds threshold	crossings at ADC value = THRES_HI and THRES_LO
--		        using pedestal value equal to adc value PED_LEAD_US upsampled points before sample XTHR_SAMP
--		        and extrapolates a straight line to cross the pedestal, the crossing point is the leading edge time returned
--		        Typical values for the external threshold, THRES_HI and THRES_LO are 5sd, 4sd, and 1sd where sd = pedestal width
--                      Typical value for XTHR_SAMP is 8, for PED_LEAD_US is 16
--
-- 	The leading edge time is returned expressed as an integer number of tenths of nanoseconds after sample 0
--
-- 	Quality codes returned are:
--              q-code: 000 : no problems
--		        001 : Constant XTHR_SAMP is too small or PED_LEAD_US is too large
--	 		010 : adc samples did not rise above pedestal + ADC_THRES_HI
--

-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;


-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

--use work.findtime_pack.all;

entity findtime is
    Port ( clk: in  STD_LOGIC;
          go: in  STD_LOGIC; -- the leading edge search process starts 1 clock sample after go is high
          reset: in std_logic;          -- reset when high
          adc: in std_logic_vector (11 downto 0);
          busy: out std_logic;          -- busy when high - several clock cycles
          completed: out std_logic;          -- done when high - one clock cycle
	  le_time : out std_logic_vector (13 downto 0); -- leading edge time found
	  q_code: out std_logic_vector (2 downto 0); -- quality code
	  state_code: out std_logic_vector (5 downto 0) -- indicator of current state (temporary)
      	  );
end findtime;

architecture Behavioral of findtime is

  
    constant NSAMPLES : integer := 15;   -- max number of ADC samples to read in
  
    type iadcarray is array (natural range <>) of integer range 0 to 4095;
     

    type task_type is (idle, fill, calcped, search, calc, calc2, calc3, calc4, calc5, calc5a, calc6, done);
    signal task : task_type := idle;


    constant XTHR_SAMP: integer := 10; -- the 5 sigma thres xing is sample[8] passed into the algo, starting with sample[0]
    constant PED_LEAD: integer := 5;  -- take local pedestal this many samples earlier, ie sample[xthr_samp-ped_lead]
    constant PED_SAMPLE : integer := XTHR_SAMP - PED_LEAD;  -- take local ped as mean of this sample and the next sample
    
    constant THRES_LO: integer := 16; -- 1 sigma
    constant SAMPLE_NS: integer := 8; -- sample period in units of 0.1ns

    constant STARTSEARCH : integer := XTHR_SAMP - PED_LEAD;  -- start search for high thres xing at this upsampled point
    


    signal iadc : iadcarray(0 to NSAMPLES-1) := (others=>0);	    -- array of adc values as integers

    signal nadc : integer range 0 to NSAMPLES := 0;  -- count of adc samples read.in


    signal adc_thres_lo : integer range 0 to 4095 := 0;

--    signal adc_sample_hi : integer range 0 to NSAMPLES*5 := 0;  --sample number for adc val over hi thres
    signal adc_sample_lo : integer range 0 to NSAMPLES*5 := 0;  -- sample num for adc val over lo thres

    signal itime : integer range 0 to NSAMPLES*10 := 0;  
    signal inum : integer range 0 to 4095 := 0;  
    signal idenom : integer range 0 to NSAMPLES := 0;  
    signal inum10 : integer range 0 to 40950 := 0; 

    signal itime5 : integer range 0 to NSAMPLES*SAMPLE_NS*10 := 0;
    
    begin

      
    process(clk)

    -- fill vars
    variable ivarped : integer range 0 to 4095 := 0;  -- diagnostic var for pedestal

    -- calc2 vars
    variable adc_val_lo : integer range 0 to 4095 := 0;

    variable sum : integer range 0 to 10 := 0;        -- for interpolation
    variable ifrac : integer range 0 to 10 := 0;
    
--    variable itime : integer range 0 to NSAMPLES*SAMPLE_NS*10 := 0;  -- ns/10
    
    begin

      if rising_edge(clk) then

        case task is

          when idle =>

            if go = '1'  then

                nadc <= 0;
                busy <= '1';
                completed <= '0';
                state_code <= "000000";
                
                task <= fill;
                
            end if;

            
          when fill => 

            nadc <= nadc + 1;
            iadc(nadc) <= to_integer(unsigned(adc));                        
          
            if nadc = NSAMPLES-1 then         --signal updates on next clock so test on previous value
              task <= calcped;
              state_code <= "000001";            
            end if;

          when calcped =>

              ivarped := (iadc(PED_SAMPLE) + iadc(PED_SAMPLE))/2;

              report "local ped " & integer'image(ivarped);
              
              --report "local ped mean of adc(" & integer'image(nadc-1) & ") and adc(" & integer'image(nadc) &  ") is " & integer'image((iadc(nadc-1) + to_integer(unsigned(adc)))/2);                           
              adc_thres_lo <= ivarped + THRES_LO;
              
              task <= search;
                     
          when search =>

            looplo: for i in XTHR_SAMP downto STARTSEARCH loop

              if iadc(i) <= adc_thres_lo then
                report "i " & integer'image(i) & " adc val " & integer'image(iadc(i)) & " lo thres " & integer'image(adc_thres_lo);
                adc_sample_lo <= i;
                exit looplo;
              end if;

            end loop looplo;

            
            task <= calc;
          

          when calc =>

            itime <= 10*adc_sample_lo;          
            task <= calc2;

          when calc2 =>

            inum <= adc_thres_lo - iadc(adc_sample_lo);
            task <= calc3;

          when calc3 =>

            idenom <= iadc(adc_sample_lo+1) - iadc(adc_sample_lo);
            task <= calc4;

          when calc4 =>

            inum10 <= inum*10;
            task <= calc5;

          when calc5 =>
            report "itime " & integer'image(itime);
            report "inum " & integer'image(inum);
            report "inum10 " & integer'image(inum10);
            report "idenom " & integer'image(idenom);

            sum := 0;
            ifrac := 0;
            loopifrac: for i in 1 to 9 loop
              sum := sum + idenom;
              ifrac := ifrac + 1;
              report "    i " & integer'image(i) & " sum " & integer'image(sum);
              exit loopifrac when sum > inum10;

            end loop loopifrac;  -- i in 0 to 10

            task <= calc5a;
            
          when calc5a =>
        
            if sum > inum10 then
              ifrac := ifrac - 1;
            elsif sum < inum10 then
              ifrac := 0;
            end if;
            
            itime5 <= itime + ifrac;
            task <= calc6;
            
          when calc6 =>
            report "ifrac " & integer'image(ifrac);
            report "itime5 " & integer'image(itime5)  & " x 0.1 samples";


--            report " " & integer'image(itime/SAMPLE_NS) & " tenths of samples since start";
--            report "sample width " & integer'image(SAMPLE_NS) & " ns";


            le_time <= std_logic_vector(to_unsigned(SAMPLE_NS*itime5,14));   -- this is time from first sample point, in 1/10ths of samples

            task <= done;

            busy <= '0';
            completed <= '1';
            q_code <= "000";
            
            state_code <= "000100";

          when done =>

            if reset='1' then
              le_time <= (others => '0');
              q_code <= (others => '0');
              completed <= '0';
              task <= idle;
              state_code <= "000110";
            end if;          
           
            state_code <= "000101";

          when others =>
            state_code <= "001000";

        end case;
      end if;

    end process;


end Behavioral;

